/***************************************************************************
 *   Copyright (C) 2006 by M.Fedotov   *
 *   anyremote@mail.ru   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include <sys/types.h>
#include <sys/wait.h>
#include <string.h>
#include <netinet/in.h>

#include <qdatetime.h>
#include <qtimer.h>

#include "logger.h"
#include "bt_comm.h"
#include "ar_client.h"
#include "ar_cmds.h"

extern logger2*  	theLog;
extern AR_Client* 	theClient;
extern AR_Parser*	conf;
extern KAnyRemote*  mainWin;

AR_Cmds::AR_Cmds() 
{
	timers 		= NULL;
	remoteOn	= true;
	resultsFile	= getToFile();        
}

AR_Cmds::~AR_Cmds() 
{
	// free timers ??
	freeTimers();
}

#include <stream.h>
#include <iostream>


QString AR_Cmds::getToFile()
{
	QString tofile("");
        char *e = getenv("TMPDIR");
        if (e) {
        	tofile += e;
        } else {
        	
        	char *h = getenv("HOME");
		if (h) {
        		tofile += h;
        		tofile += "/.anyRemote";
		} else {
			tofile += "/tmp";
		}
        }
        tofile += TOFILE;
        char *u = getenv("USER");
        if (e && u) {
        	tofile += QString(".") + u;
        }  
	//cout << tofile  << endl; 
        return tofile;
}

void AR_Cmds::setStyle(bool isServer)
{
	isServerMode = isServer;
}

void AR_Cmds::setRemoteOn(bool set)
{
	theLog->info(QString("Switch remote ") + (set ? "On" : "Off"));
	remoteOn = set;
}

bool AR_Cmds::isRemoteOn(void)
{
	return remoteOn;
}

int AR_Cmds::execInFork(const char *cmd)
{
	int ret;
	int cpid = fork();
	if( cpid < 0) {
		theLog->error("Can not fork!");
		return -1;
	}

	if(cpid) {	// father
		waitpid(cpid,&ret,0);
		if(ret!=-1 && WIFEXITED(ret)) {
			ret = WEXITSTATUS(ret);
		}
	} else {	// child
		theLog->info("Close port in child");
		theClient->getComm().closePort(0);	
		ret = system(cmd);
		if(ret!=-1 && WIFEXITED(ret)) {
			ret = WEXITSTATUS(ret);
		}
		exit(ret);
	}

	return ret;
}

char * AR_Cmds::readResultsFile(int * count) 
{
	FILE *fp;
	struct stat buf;
	
	if(stat(resultsFile.ascii(), &buf) == -1) {
		return NULL;
	}
	
	long fLen = buf.st_size;
	theLog->info("readResultsFile >" + QString().setNum(fLen) + "<");
	
	fp=fopen(resultsFile.ascii(),"r");
	if (fp == NULL) {
		return NULL;
	}
    
    	char* fBuffer = (char*) calloc(fLen+1,1);
    
	(*count) = fread(fBuffer, sizeof(char), fLen, fp);
	fclose(fp);
        
	return fBuffer;
}

QString AR_Cmds::substParams(QString command, cmdParams *params)
{
	if (command.isEmpty()) { // Empty command. Skip it.
		theLog->error("substParams(): null input");
		return QString::null;
	}
	theLog->info("substParams >" + command + "<");

	int papameterized = 0;
	QString cmd("");

	while(!command.isEmpty()) {
		
		papameterized = command.find(PARAM_START);   // $(

		if (papameterized < 0) {	    
			theLog->debug("non-parametrized command (1)");
			cmd += command;
			break;
		} else if (command.mid(papameterized+2).startsWith(PARAM_CALLID)) {	    // Need to substitute it with callerID
		    
			theLog->debug("parametrized command - callerID");
		
			// First part
			cmd += command.left(papameterized);
	
			// Add params and second part
			cmd += theClient->getCallerID();
	
			command = command.right(command.length() - papameterized - QString(PARAM_CALLID).length() - 2);

		} else if (command.mid(papameterized+2).startsWith(PARAM_TIME)) {     // Need to substitute it with time
		    
			theLog->debug("parametrized command - Time");

			// First part
			cmd += command.left(papameterized);
		
			QString time = QTime::currentTime().toString().left(5);
		
			// Add params and second part
			cmd += time;
	
			command = command.right(command.length() - papameterized - QString(PARAM_TIME).length() - 2);
	
		} else if (command.mid(papameterized+2).startsWith(PARAM_MODE)) {   // Need to substitute it with mode
	
			theLog->debug("parametrized command - Mode");
		
			// First part
			cmd += command.left(papameterized);
		
			// Add params and second part
			cmd += conf->getMode();
	
			command = command.right(command.length() - papameterized - QString(PARAM_MODE).length() - 2);	    

		} else if (command.mid(papameterized+2).startsWith(PARAM_PARAM)) {
			// This is dirty trick to make Bemused commands like VOLM<volume> works
			// also it is used in Java Client list handling
		
			if (params == NULL) {
				theLog->debug("Parameter value was not set!");
                                cmd += QString(PARAM_START) + PARAM_PARAM;
			} else {
				// First part
				cmd += command.left(papameterized);
		
				// Add params
				cmd += params->value;
			}
			command = command.right(command.length() - papameterized - QString(PARAM_PARAM).length() - 2);      

		} else if (command.mid(papameterized+2).startsWith(PARAM_INDEX)) {
			// Used in Java Client list handling

			if (params == NULL || params->index.isEmpty()) {
				theLog->debug("Parameter value was not set!");
                                cmd += QString(PARAM_START) + PARAM_INDEX;
			} else {
				// First part
				cmd += command.left(papameterized);
		
				// Add params
				cmd += params->index;
			}
			command = command.right(command.length() - papameterized - QString(PARAM_INDEX).length() - 2);      

		} else if (command.mid(papameterized+2).startsWith(PARAM_CFGDIR)) {
		    
			theLog->debug("parametrized command - CfgDir");
		
			// First part
			cmd += command.left(papameterized);
	
			// Add params and second part
			cmd += conf->getCfgDir();
	
			command = command.right(command.length() - papameterized - QString(PARAM_CFGDIR).length() - 2);
	
		} else {    // Just copy
			
			theLog->debug("no more parameters");
			cmd += command;
			break;
		}
	}

	theLog->debug("final cmd >" + cmd + "<");
	return cmd;
}

int AR_Cmds::execSimpleCmd(QString command, int toFile)
{
	if (command.isEmpty()) { // Empty command. Skip it.
		theLog->info("execSimpleCmd null input");
		return EXIT_OK;
	}
	theLog->info("execSimpleCmd >" + command + "<");

	if (!isRemoteOn()) {
		theLog->debug("Remote if OFF. Skip event");
		return EXIT_OK;
	}

	if (toFile) {	    // We needs to wait results, so do not run in background
		command += QString(" > ") + resultsFile;
	} else {
		if (!command.contains("&")) {
			command += QString(" &");
		}
	}

	return (execInFork(command.ascii()) == -1 ? EXIT_NOK : EXIT_OK);
}

int AR_Cmds::processSendCmd(cmdItem* ci)
{
	int ret = EXIT_OK;

	QString tag  = ci->descr.section(",",0,0).simplifyWhiteSpace();
	QString cmd  = ci->descr.section(",",1);  // no simplifyWhiteSpace()

	if (tag.isNull() || tag.isEmpty() || cmd.isNull() || cmd.isEmpty()) {
		theLog->debug("processSendCmd(): incorrect input (1)"); 	       
		return EXIT_NOK;
	}

	theLog->debug("processSendCmd() is >" + tag + "< >" + cmd + "<");

	if (tag == SEND_STRING) {
		ret = theClient->getComm().writeSocketConn((char *)cmd.ascii(), strlen(cmd.ascii()));
		if (ret != EXIT_OK) {
		 	theLog->debug("Fails in writeSocketConn()");
		}
	} else if (tag == SEND_BYTES) {
		ret = theClient->getComm().writeBytesSocketConn((char *)cmd.ascii());
		if (ret != EXIT_OK) {
			theLog->debug("Fails in writeBytesSocketConn()");
		}    
	} else {
		theLog->debug("processSendCmd(): incorrect input (3)"); 	       
		return EXIT_NOK;
	}
	return ret;
}

void AR_Cmds::sendData(int type, bool bytesFlag, char *data, int count) 
{
	int ret;
	if (type == ID_SEND && bytesFlag) {
		ret = theClient->getComm().writeBytesSocketConn(data);
	} else {
		ret = theClient->getComm().writeSocketConn(data, count);
	}
	if (ret != EXIT_OK) {
		theLog->debug("Fails in write(Bytes)SocketConn()");
	}
}

int AR_Cmds::macroCmd(QString macro, cmdParams *params) 
{
 	theLog->debug("macroCmd");
        
	int ret;
	type_key *k = conf->findItem(macro, ret, NULL);

	if (k && ret == FLAG_EXACT) {
		ret = handleCmdByKey(k, params);
		if (ret == EXIT_ABORT || ret == EXIT_DISCON) {
			return ret;
		}
	}
	return EXIT_OK;
}

int AR_Cmds::loadCmds(QString file) 
{
	theLog->debug("loadCmds");
        
	FILE *fp;
        char *command = NULL;
        size_t len = 0;
        ssize_t read;
        
        if (file.isEmpty()) {
        	return EXIT_NOK;
        }

	fp=fopen(file.ascii(),"r");
	if (fp == NULL) {
		return EXIT_NOK;
	}
        
	while ((read = getline(&command, &len, fp)) != -1) {
        	theLog->debug(QString("Retrieved line ") + command);
                
        	cmdItem* tmpCmd = NULL;
                conf->storeCmds(&tmpCmd, command); // parse command on-the-fly 
                
                processCommands(tmpCmd,NULL);
                
                conf->freeCmds(tmpCmd);
        }
        if (command) {
        	free(command);
        }
        fclose(fp);
	return EXIT_OK;
}

int AR_Cmds::includeCmd(QString file) 
{
	theLog->debug("includeCmd");
        
        if (file.isEmpty()) {
		theLog->debug("File empty");
		return EXIT_NOK;
        }

        conf->load_cfg(file, false);
        //conf->printKeys();
        conf->printConf();
        
	return EXIT_OK;
}

void AR_Cmds::uploadPix(char* cmdLine)
{
        theLog->debug(QString("uploadPix >")+cmdLine+"<");
	int prefixSz = 0;          		// image,icon,_name_, | _file_) or image,pix, | _file_)
        
        if (strstr(cmdLine,UPLOAD_PIX) != NULL) {
        	prefixSz = strlen(UPLOAD_PIX) + 1;
        } else {
        	int nCh = strlen(UPLOAD_ICON) + 1;
       		prefixSz = nCh + strstr(cmdLine+nCh,",") - (cmdLine+nCh) + 1;
	}
        
	char* fname = cmdLine + prefixSz;
        char* p = fname;
        while (*p != '\0') {	// remove trailing \n if exists
        	if (*p == '\n') {
                	*p = '\0';
                        break;
                }
                p++;
        }
        theLog->debug(QString("uploadPix file >")+fname+"<"); 

	FILE *fp;
	struct stat buf;
	
        if(stat(fname, &buf) == -1) {
        	theLog->error("can't get file size!");
		return;
	}
	
	long fLen = buf.st_size;
	
        fp=fopen(fname,"r");
	if (fp == NULL) {
        	theLog->error("can't open file!");
		return;
	}
        
    	uint32_t szh32 = (uint32_t) fLen;
    	uint32_t szi32 = htonl(szh32);
    
    	char * fBuffer = NULL;
    	fBuffer = (char *) calloc(fLen+prefixSz+12,1);	// 11 =  "Set("     ,<size of bin data=4bytes>    ");\0"
        if (fBuffer == NULL) {
        	theLog->error("no more memory!");
		return;
        }
        
    	strcpy(fBuffer, "Set(");
    	strncat(fBuffer, cmdLine, prefixSz+1);
        memcpy((void*)(fBuffer+prefixSz+4), (const void *) &szi32, 4);	// length on binary data
        
	fread(fBuffer+prefixSz+8, sizeof(char), fLen, fp);
    	fclose(fp);

        strcpy(fBuffer+prefixSz+fLen+8,");");
        
	int ret = theClient->getComm().writeSocketConn(fBuffer,prefixSz+fLen+10);
	if (ret != EXIT_OK) {
		theLog->error("uploadPix(): Fails in writeSocketConn()");
	}

        free(fBuffer);
	
        return;
}

int AR_Cmds::execCmdAndGetResults(int type, cmdItem* ci, timerCmd* timer, cmdParams *params) 
{
	theLog->debug("execCmdAndGetResults");

	if (!isServerMode || !isRemoteOn() || ci->exec.isEmpty()) {
        	return EXIT_OK;
	}
        
	QString sDescr = substParams(ci->descr, params);
	if (sDescr.isEmpty()) {
        	theLog->debug("Empty description after substParams()");
        }
	QString sExec = substParams(ci->exec, params);
	if (sExec.isEmpty()) {
        	theLog->debug("Empty exec. part after substParams()");
        }
        
        int ret = execSimpleCmd(sExec,1);
        if (ret != EXIT_OK) {
        	theLog->debug("Return code from execSimpleCmd() is " + QString().setNum(ret));
        	return ret;
        } 
           
 	int resLen = -1;
        char *res  = readResultsFile(&resLen);
	
	if (res == NULL) {
		theLog->error("readResultsFile() returns NULL");
		return EXIT_NOK;
	}

        theLog->debug(QString("readResultsFile() ") + res);

        if (timer != NULL) {
        	if (timer->lastValue != NULL && 
                    memcmp(timer->lastValue,res,timer->lastValueSize) == 0) {
                    	free(res);	
        		return EXIT_OK;
                } else {
                	updateTimerLastData(timer, res, resLen);		// Update "last data"
                }
        }

        if (type == ID_SEND) {  // Just send string
        
               	bool bytesFlag = false;
                if (sDescr.contains(SEND_BYTES)) {
                	bytesFlag = true;
                }
        	sendData(type, bytesFlag, res, resLen);
                
        } else {  		// ID_SET

                if (sDescr.contains(UPLOAD_ICON) || 
                    sDescr.contains(UPLOAD_PIX)) {
                        
                        char *chunk = (char *) calloc(strlen(sDescr.ascii())+resLen+2,1);
                        
                        strcpy(chunk,sDescr.ascii());
                        strcat(chunk,",");
                        memcpy(chunk+strlen(sDescr.ascii())+1,res,resLen);
                       
                	uploadPix(chunk);
                        
                        free (chunk);
     
                } else {
                	// Replace all ");" inside buffer to "@@" since Java Client will separate commands by ");"
                	char *p;
                	while ((p = strstr(res,");")) != NULL) {	// binary data ???
                        	*p = '@';
                        	p++;
                        	*p = '@';
                	}

                	int sz = strlen(sDescr.ascii());
                
                	int memSize = resLen + sz + 7;
                	char *chunk = (char *) calloc(1,memSize+1);
                	strcpy(chunk,CMD_SET);

                	strcat(chunk,"(");  // For example, Set(title
                	strcat(chunk,sDescr.ascii());
                
                	strcat(chunk,",");
                	memcpy(chunk+5+sz,res,resLen);	// could be binary data here
                	strcpy(chunk+5+sz+resLen,");");
                	
                	//printf("CHUNK: %s\n", chunk);
                
                	sendData(type, 0, chunk, memSize);
                
                	free (chunk);
                }
        }
        
	free(res);
        
	return EXIT_OK;
}

int AR_Cmds::setMode(QString& mode) 
{
	int ret;
	int i = 0;    
        type_key* k = conf->findItem(QString(EVT_EXITMODE), i, NULL);
        if (k != NULL && i == FLAG_EXACT) {
        	theLog->debug("Exec cmd on exit from mode");
        	ret = handleCmdByKey(k, NULL);
        	if (ret == EXIT_ABORT || ret == EXIT_DISCON) {
        		return ret;
        	}
        } else {
        	theLog->debug("Leave mode\n");
        }
        conf->setMode(mode);

        i = 0; 
        k = conf->findItem(QString(EVT_ENTRMODE), i, NULL);
        if (k != NULL && i == FLAG_EXACT) {
        	theLog->debug("Exec cmd on enter to mode");
        	ret = handleCmdByKey(k, NULL);
        	if (ret == EXIT_ABORT || ret == EXIT_DISCON) {
        		return ret;
        	}
        } else {
        	theLog->debug("Entered to mode");
        }
	return EXIT_OK;
}

int AR_Cmds::processOneCommand(cmdItem* ci, cmdParams *params)
{
	int ret;

	if (ci->descr.isEmpty() || ci->exec.isEmpty()) {
		theLog->info("execCmd >no description or exec params<");
	} else {
		theLog->info("execCmd >" + ci->descr + "< >" + ci->exec + "<");
	}    
         
	if (ci->type == ID_REMOTEON) {
                theLog->info("Switch remote on");
  		setRemoteOn(true);      
                return EXIT_OK;
        }
        if (!isRemoteOn()) {
               theLog->info("Remote is off. Skip");
               return EXIT_OK;
        }
        
	switch (ci->type) {	// Non parametrized commands
		
		case ID_EXIT:		    
			theLog->info("Exit key detected");			// User asks to exit	
			theClient->Disconnect();			    
			return EXIT_ABORT;
	
		case ID_TMPDISCONN:
			theLog->info("Temporary disconnect key detected");	// User asks to temporary disconnect
			theClient->Disconnect();
			return EXIT_DISCON; // But what with the rest of the command line ? (now just skip)
	
		case ID_REMOTEOFF:
			theLog->info("Switch remote off");
			setRemoteOn(false);
			return EXIT_OK;
                
		case ID_TIMERCMD:   
			theLog->info("Command: Timer");
			if (isServerMode) {
				return processTimerCmd(ci);
			}     
			return EXIT_OK;

		case ID_CANCELTIMER:   
			theLog->info("Command: CancelTimer");
			if (isServerMode) {
				cancelTimer(ci->descr);
			}  
			return EXIT_OK;

		case ID_SEND:
			theLog->info("Command: Send");
			if (isServerMode) {
				ret = processSendCmd(ci);
				if (ret != EXIT_OK) {
					theLog->debug("Fails in processSendCmd()");
				}
			}    
			return EXIT_OK;

                default:
                	break;        
        }
                    
	switch (ci->type) {		// Parametrized commands, will parametrize in handler
	
		
		case ID_EXECSET:   
			theLog->info("Command: ExecAndSet");
			execCmdAndGetResults(ID_SET, ci, NULL, params);
			return EXIT_OK; 
	
		case ID_EXECSEND:   
			theLog->info("Command: ExecAndSend");
			execCmdAndGetResults(ID_SEND, ci, NULL, params);
			return EXIT_OK;
		   
                case ID_MACRO:   
			theLog->info("Command: Macro");
			ret = macroCmd(ci->descr, params);
              		if (ret == EXIT_ABORT || ret == EXIT_DISCON) {
                        	return ret;
                        }
			return EXIT_OK;

		default:  
			break;
	}
        
        if (ci->type == ID_EXEC) {	// Parametrized command, handler will use already parametrized data
		QString sExec = substParams(ci->exec, params);
		if (sExec.isEmpty()) {
        		theLog->debug("Empty exec. part after substParams()");
        	}
		theLog->info("Command: Exec");
	        execSimpleCmd(sExec, 0);
	        return EXIT_OK;
        }
        
	QString sDescr = substParams(ci->descr, params);
        if (sDescr.isEmpty()) {
        	theLog->debug("Empty description after substParams()");
        }
        
	switch (ci->type) {		// Parametrized commands, handler will use already parametrized data
	
		case ID_SETMODE:
			theLog->info("Command: SetMode");
                        
			ret = setMode(sDescr);
                        
			if (ret == EXIT_ABORT || ret == EXIT_DISCON) {
		            	return ret;
		        }
			break;

		case ID_SENDCKPD:
			theLog->info("Command: SendCKPD");
			if (!isServerMode) {
				ret = theClient->getComm().sendSeq(sDescr);
				if (ret != EXIT_OK) {
					theLog->debug("Fails in sendSeq()");
				}

				if (ret == EXIT_ABORT || ret == EXIT_DISCON) {
					return ret;
				}
			}    
			break;

		case ID_SET:	      
			theLog->info("Command: Set");
			if (isServerMode && ! sDescr.isEmpty()) {

				QString tmptext(CMD_SET);
				tmptext += QString("(");
		
                 		if (strstr(sDescr,UPLOAD_ICON) != NULL || 	// handle Set(image,icon|window,..._image_file_name_); separately.
                    		    strstr(sDescr,UPLOAD_PIX)  != NULL) {
                                    
				    char * dsc = strdup(sDescr.ascii());
				    
                                    uploadPix(dsc);
				    
				    free(dsc);

				} else {
					int i = 0;
					while (true) {
			               		QString part = sDescr.section("\\n",i,i);
			               		if (part.isNull()) {
			                       		break;
			               		}
			               		tmptext += part + "\n";
			       			i++;
					}
					tmptext += ");";
			
					ret = theClient->getComm().writeSocketConn((char *)tmptext.ascii(), strlen(tmptext.ascii()));
					if (ret != EXIT_OK) {
						theLog->debug("Fails in writeSocketConn() (Set)");
					}
                                }
 			}    
			break;
		
             	case ID_GET:          
			theLog->info("Command: Get");
                	if (isServerMode && ! sDescr.isEmpty()) {
				QString tmptext(CMD_GET);

                        	tmptext += "(" + sDescr + ");";
			
                        	ret = theClient->getComm().writeSocketConn((char *)tmptext.ascii(), strlen(tmptext.ascii()));
                        	if (ret != EXIT_OK) {
                         		theLog->debug("Fails in writeSocketConn() (Get)");
                         	}
            		}    
             		break;
                
		case ID_LOAD:   
			theLog->info("Command: Load");
			ret = loadCmds(sDescr);
              		if (ret == EXIT_ABORT || ret == EXIT_DISCON) {
                        	return ret;
                        }
			break;

		case ID_INCLUDE:   
			theLog->info("Command: Include");
			ret = includeCmd(sDescr);
              		if (ret == EXIT_ABORT || ret == EXIT_DISCON) {
                        	return ret;
                        }
			break;

		default:  
			theLog->info("Unknown command");		 
			break;
	}

	return EXIT_OK;
}

int AR_Cmds::processCommands(cmdItem* cmdList, cmdParams *params) 
{
	while (cmdList) {
		
		int ret = processOneCommand(cmdList, params);
		if (ret != EXIT_OK) {
			theLog->debug("Return code from processOneCommand() is " + QString().setNum(ret));
			 
			if (ret == EXIT_ABORT || ret == EXIT_DISCON) {
				theLog->debug("Clean auto repeat flag after processOneCommand()");

				theClient->cleanupRepeatKeys();
		
				return ret;
			}
		}
		cmdList = cmdList->next;
	}

	return EXIT_OK;
}

//
// Top-level command processing function
//

int AR_Cmds::handleCmdByKey(type_key *tk, cmdParams *params)
{
	if(tk == NULL) {
		theLog->debug("handleCmdByKey() NULL input");
		return EXIT_OK;
	}
	theLog->debug("handleCmdByKey() >" + tk->key + "<");

	cmdItem* cmdList = conf->getCommand(tk);
	if (cmdList == NULL) {
		theLog->debug("No appropriate command was found by getCommand()");
	} else {
		theLog->debug("Ready to execute command");
	
	
		int ret = processCommands(cmdList, params);
		
		// In AT mode: User had pressed some key ... 
		// Send ToMainMenu sequence of CKPD's to show main menu again
		if (isRemoteOn() && !isServerMode) {
			if (theClient->getComm().sendSeq(conf->getToMainMenu()) != EXIT_OK) {
				theLog->debug("Fails in sendSeq()");
			}
		}
		
		return ret;
	}   

	return EXIT_OK;     
}

//
// Functions related to timer commands
//
int AR_Cmds::processTimerCmd(cmdItem* ci)
{
	if (ci->exec.isEmpty() || ci->tdata == NULL) {
        	theLog->debug("processTimerCmd(): nothing to exec or timer data absent");
                return EXIT_NOK;
        }
        
	int ret = EXIT_OK;

	theLog->debug("processTimerCmd(): >" + ci->exec + "<");

	timerCmd* tc = NULL;
	if ((tc=createTimer(ci)) == NULL) {
		theLog->debug("processTimerCmd(): timer already exists or input was incorrect");		
		return EXIT_NOK;
	}

	execCmdAndGetResults(ID_SET, ci, tc, NULL);

	// AR_Cmds is not successor of QObject so send timeout to mainWin
	theLog->debug("Set commands timer");
	QTimer::singleShot(500, mainWin, SLOT(handleTimerCmd()));

	return ret;
}

timerCmd* AR_Cmds::findTimer(QString& id, timerCmd **prev) 
{
	if (id.isEmpty()) {
        	return NULL;
        }
	timerCmd* ptr = timers;

	while (ptr != NULL && (!ptr->timer->tdata->id.startsWith(id))) {
        	if (prev != NULL) {
			*prev = ptr;
			theLog->debug("findTimer(): set prev to >" + (*prev)->timer->tdata->id + "<");
                }
		ptr  = ptr->next;
	}
	return ptr;
}

timerCmd* AR_Cmds::createTimer(cmdItem *ci)
{
	if (ci->tdata->id.isEmpty() || ci->exec.isEmpty()) {
		theLog->debug("createTimer(): wrong input");		   
		return NULL;
	}

	if (findTimer(ci->tdata->id, NULL) != NULL) {
		theLog->debug("createTimer(): timer already exists");		     
		return NULL;
	} 
	timerCmd *tm = new timerCmd();

	tm->ticks     = 0;
	tm->times     = 1;
        tm->timer     = ci;
        tm->lastValue = NULL;
        tm->lastValueSize = -1;

	// Insert in head
	if (timers == NULL) { // Insert first
		tm->next = NULL;
	} else {
		tm->next = timers;
	}
	timers = tm;

	return tm;
}

int AR_Cmds::cancelTimer(QString& id) 
{
	theLog->debug("cancelTimer() >" + id + "<");

	timerCmd *prev = NULL;
	timerCmd *tm = findTimer(id, (timerCmd **) &prev);
	if (tm == NULL) {
		theLog->debug("Cant find timer to cancel!");
		return EXIT_NOK;
	}    
	if (tm == timers) { //remove in head
		timers = tm->next;
	} else {
		if (prev != NULL) {
			prev->next = tm->next;
		} else {
			theLog->error("Previous item absent for non-first element");
			return EXIT_NOK;
		}
	}	 
	delete tm;

	return EXIT_OK;
}

void AR_Cmds::freeTimers()
{
	//theLog->debug("freeTimers()");
	timerCmd *tm ;

	while (timers != NULL) {
		tm     = timers;
		timers = timers  ->next;

                tm->timer = NULL;

		free(tm);
	}
	timers = NULL;
	return;
}

int AR_Cmds::verifyTimers(int ticksInSec) 
{
	timerCmd *tm   = timers;
	int numTimers  = 0;
        
	while (tm != NULL) {
		
		if (tm->ticks >= (tm->timer->tdata->timeout * ticksInSec)) {
			theLog->debug("verifyTimers(): it is time to execute >" + tm->timer->exec + "<");

			execCmdAndGetResults(ID_SET, tm->timer, tm, NULL);
	
			tm->ticks = 0;
	
			if (tm->timer->tdata->times > 0) {
				tm->times++;
				if (tm->times >= tm->timer->tdata->times) {    // Cancel this timer 
					theLog->debug("verifyTimers(): timer " + tm->timer->tdata->id + " self-canceled");
			
					// Cancel this timer
					timerCmd *tmc = tm;
					tm = tm->next;

					cancelTimer(tmc->timer->tdata->id);
			
					continue;
				}
			}
		} else {
			tm->ticks++;
		}
		//prev = tm;
		tm   = tm->next;
		numTimers++;
	}
	return numTimers;
}

void AR_Cmds::updateTimerLastData (timerCmd* timer, char * data, int size)
{
	if (timer->lastValue != NULL) {
        	free(timer->lastValue);
                timer->lastValue = NULL;
        }
        
        timer->lastValue     = data;
        timer->lastValue     = (char *) calloc(size + 1, 1);
        timer->lastValueSize = size;
        
        memcpy(timer->lastValue,data,size);
}
