/***************************************************************************
 *   Copyright (C) 2006 by M.Fedotov   *
 *   anyremote@mail.ru   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#ifndef AR_CMDS_H
#define AR_CMDS_H

#include "ar_parser.h"

#define PARAM_START     "$("
#define PARAM_CALLID    "CallId)"
#define PARAM_TIME      "Time)"
#define PARAM_MODE      "Mode)"
#define PARAM_PARAM     "Param)"
#define PARAM_INDEX     "Index)"
#define PARAM_CFGDIR    "CfgDir)"

/**
	@author M.Fedotov <anyremote@mail.ru>
*/


typedef struct TC
{
	cmdItem *timer;
        
	int	ticks;
	int	times;
	char*	lastValue;
	int     lastValueSize;        
	struct TC  *next;
} timerCmd;

class AR_Cmds {

public:
	AR_Cmds();
	~AR_Cmds();

	int handleCmdByKey	(type_key *tk, cmdParams *params);
	int  processCommands	(cmdItem* cmdList, cmdParams *p);
	void setStyle		(bool isServer);
	QString getToFile       (void);

	void freeTimers         (void);
	int  verifyTimers       (int ticksInSec);
	int  cancelTimer        (QString& id);
	int  processTimerCmd    (cmdItem* ci);
	void setRemoteOn	(bool set);
	bool isRemoteOn	    	(void);
	void sendData		(int type, bool byteFlag, char* data, int count);
	int setMode		(QString& mode);
	int execInFork  	(const char *cmd);

private:

	int processOneCommand	(cmdItem* ci, cmdParams *params);
	int execSimpleCmd	(QString command, int toFile);
	int processSendCmd	(cmdItem* ci);
	char* readResultsFile   (int* l);  
	int macroCmd            (QString macro, cmdParams *params);
	int loadCmds            (QString f);
	int includeCmd		(QString f);
        int execCmdAndGetResults(int type, cmdItem* ci, timerCmd *tm, cmdParams *params);
	timerCmd* findTimer	(QString& id, timerCmd **prev);
	timerCmd* createTimer   (cmdItem* ci);
	void updateTimerLastData(timerCmd* timer, char* data, int size);
	QString substParams     (QString command, cmdParams *params);
	void uploadPix		(char *cmdLine);
        
	int 		isServerMode;
	int 		remoteOn;
	timerCmd* 	timers;
        
        QString         resultsFile;
};

#endif
