/***************************************************************************
 *   Copyright (C) 2006 by M.Fedotov   *
 *   anyremote@mail.ru   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include <signal.h>
#include <dirent.h>

#include <qtooltip.h>
#include <qtimer.h>
#include <qlineedit.h>
#include <qpixmap.h>
#include <qfile.h>
#include <qcheckbox.h>

#include <kaction.h>
#include <kapplication.h>
#if !(KDE_VERSION_MINOR<=1 && KDE_VERSION_RELEASE<=5)
#include <kactioncollection.h>
#endif
#include <kcmdlineargs.h>
#include <kdebug.h>
#include <khelpmenu.h>
#include <kiconloader.h>
#include <klocale.h>
#include <kmainwindow.h>
#include <kmessagebox.h>
#include <kpassivepopup.h>
#include <kpopupmenu.h>
#include <ksimpleconfig.h>
#include <kurlrequester.h>

#include "kanyremote.h"
#include "kanyremotewidget.h"
#include "ar_parser.h"
#include "ar_client.h"
#include "ar_cmds.h"
#include "ar_dcop.h"

#include "logger.h"

extern bool 	    printLog;
extern QString	    serial;
extern QString	    cfgFile;

KAnyRemoteWidget  *cfgWin;
AR_Parser *conf 	 = NULL;
AR_Cmds   *commander = NULL;
AR_Client* theClient = NULL;
logger2   *theLog	 = NULL;
AR_Editor *theEditor	= NULL;
AR_dcop	  *theDcop	= NULL;
QPixmap   *On 		= NULL;
QPixmap   *OnBig 	= NULL;
QPixmap   *Off 		= NULL;
QPixmap   *Flash 	= NULL;

extern void logger(char *head, const char *str);
void KAnyRemoteTrayIcon::mousePressEvent(QMouseEvent *e)
{
	KSystemTray::mousePressEvent(new QMouseEvent(QEvent::MouseButtonPress, e->pos(), e->globalPos(), e->button() == LeftButton ? RightButton : e->button(), e->state()));
}

QPixmap* KAnyRemote::getPixmapByName(QString name)
{
	// search icons in
	// /usr/share/pixmaps/kanyremote
	// /usr/local/share/pixmaps/kanyremote
	// /usr/local/kde/share/pixmaps/kanyremote
	// ./src/pix
	// .
	QFile pixmap;
	QPixmap* Px = NULL;

	QString fname = QString("/usr/share/pixmaps/kanyremote/") + name + ".png";
	theLog->info(QString("Try QPixmap ") + fname);
	pixmap.setName(fname);
	if (pixmap.exists()) {
		Px = new QPixmap(fname);
		return Px;
	}
	
	fname = QString("/usr/local/share/pixmaps/kanyremote/") + name + ".png";
	theLog->info(QString("Try QPixmap ") + fname);
	pixmap.setName(fname);
	if (pixmap.exists()) {
		Px = new QPixmap(fname);
		return Px;
	}

	fname = QString("/usr/local/kde/share/pixmaps/kanyremote/") + name + ".png";
	theLog->info(QString("Try QPixmap ") + fname);
	pixmap.setName(fname);
	if (pixmap.exists()) {
		Px = new QPixmap(fname);
		return Px;
	}
		
	fname = "./src/pix/" + name + ".png";
	theLog->info(QString("Try QPixmap ") + fname);
	pixmap.setName(fname);
	if (pixmap.exists()) {
		Px = new QPixmap(fname);
		return Px;
	} 

	fname = "./" + name + ".png";
	theLog->info(QString("Try QPixmap ") + fname);
	pixmap.setName(fname);
	if (pixmap.exists()) {
		Px = new QPixmap(fname);
		return Px;
	} 
	
	theLog->info("Try QPixmap SmallIcon(" + name+ ")");
	Px = new QPixmap(SmallIcon(name));
	return Px;
}

void KAnyRemote::setPixmapsAndTooltips(int mode, QString newMessage)
{
	QPixmap *px = NULL;
	QString msg;
	QString menu;

	if (mode == 0) {    	    	    	// Off
		px 		= Off;
		msg 	= "Unconnected";
		menu 	= "&Connect...";
	} else if (mode == 1) {  		// On
		px 		= On;
		msg 	= "Ready";
		menu 	= "&Disconnect...";
	} 

	if (! newMessage.isNull()) {
		KPassivePopup::message("KAnyRemote", i18n(newMessage), *On, theTrayIcon);
	}
    	if (!theClient->isServer()) {
	    theTrayIcon->contextMenu()->changeItem (menuItemConnectID, i18n(menu));
	}
	theTrayIcon->setPixmap(*px);
	QToolTip::add(theTrayIcon, i18n(msg));
}

KAnyRemote::KAnyRemote() : KMainWindow( 0, "KAnyRemote" )
{
        // Not to core dump if connection will close
        signal(SIGPIPE, SIG_IGN); 
	
	state = STATE_NOTINITED;
	
        setCentralWidget((cfgWin = new KAnyRemoteWidget(this)));

	// Create all suplimentaries
	theTrayIcon 	= new KAnyRemoteTrayIcon();
	conf 		= new AR_Parser();
	theLog		= new logger2();
	theClient 	= new AR_Client();
	commander 	= new AR_Cmds();
	theDcop 	= new AR_dcop();
	theResetCount 	= 0;
	theRepeatCounter = 0;

	// Set name of cfg file if any
	loadFromConfig();

	// Override with -f ... option
	if (!(cfgFile.isNull() || cfgFile.isEmpty())) {
		//theLog->info(QString("Override cfgFileName=") + cfgFile);
		cfgFileName = cfgFile;
	}
	
	// Load data from cfg file
	resetCfg(false);

	// Override -log
	if (printLog) {
		//theLog->info("Override log=true");
		conf->setLog (printLog);
		theLog->block(! printLog);
		theLog->setFile(conf->getLog(), conf->getDebug());
	}

	// Override -serial ...
	if (!serial.isNull()) {
		//theLog->info("Override device=" + serial);
		conf->setDevice(serial);
	}
	
	theLog->cfg("---------------- Configuration info ----------------"); 
	conf->printConf();
	
	// Determine is this server or client style and parse port/channel
	theClient->setStyle(conf->getDevice());
	commander->setStyle(theClient->isServer());

	state = STATE_DEFSTYLE;

	// GUI stuff
	theFlashOff = new QTimer(theTrayIcon);

	connect(theFlashOff, SIGNAL(timeout()), SLOT(flashOff()));

	// Search pixmaps in: .:$HOME:
	On    = getPixmapByName("kanyremote");
	Off   = getPixmapByName("kanyremote_off");
	Flash = getPixmapByName("kanyremote_flash");

	//Seems changeTitle() does notworks ...No idea why ...
	theTrayIcon->contextMenu()->changeTitle(0, QPixmap(*On), "KAnyRemote");
	theTrayIcon->contextMenu()->insertItem(SmallIcon( "configure" ), i18n("&Configuration"), this, SLOT(slotConfigure()));
	
	menuItemConnectID = theTrayIcon->contextMenu()->insertItem(SmallIcon( "connect_creating" ), i18n("&Connect..."), this, SLOT(slotConnect()));
	theTrayIcon->contextMenu()->insertSeparator();
	theTrayIcon->contextMenu()->setItemVisible (menuItemConnectID, (!theClient->isServer()));
	
	//theTrayIcon->contextMenu()->insertItem(SmallIcon( "help" ), KStdGuiItem::help().text(), 
	//             (new KHelpMenu(theTrayIcon, KGlobal::instance()->aboutData(),false))->menu());

	About = new KHelpMenu(theTrayIcon, KGlobal::instance()->aboutData(),false);
	theTrayIcon->contextMenu()->insertItem(SmallIcon( "help" ), KStdGuiItem::help().text(), this, SLOT(showAbout()));
	
	theTrayIcon->actionCollection()->action("file_quit")->disconnect(SIGNAL(activated()));
	connect(theTrayIcon->actionCollection()->action("file_quit"), SIGNAL(activated()), SLOT(doQuit()));
    		
	if (conf->getAutoConnect() || theClient->isServer()) {
		//checkBluetooth(); to get time to draw icon in toolbar - run connect after 1 sec
		
		QTimer::singleShot(1000, this, SLOT(checkBluetooth()));
		setPixmapsAndTooltips(0);
	} else {
		setPixmapsAndTooltips(0);
	}
	theTrayIcon->show();
}

KAnyRemote::~KAnyRemote()
{
	delete theTrayIcon;
	delete theClient;
	delete commander;
	delete theLog;
	delete conf;
	delete On;
	delete OnBig;
	delete Off;
	delete Flash;
	delete About;
}

void KAnyRemote::createDataDir() 
{
	char dd[542];
	char *t = getenv("HOME");
	strcpy(dd, t);
	strcat(dd, "/.anyRemote");

	DIR *d = opendir(dd);
	if (d) {
        	closedir(d);
        } else {
		mkdir(dd, S_IRWXU);
	}
}

void KAnyRemote::gotMessage(const QString &theButton)
{
	theLog->debug("KAnyRemote::gotMessage()");
	theLog->debug(QString("Event: ") + theButton);

	theTrayIcon->setPixmap(*Flash);
	theFlashOff->start(200, true);
}

void KAnyRemote::flashOff()
{	
	theLog->debug("KAnyRemote::flashOff()");
	if (theClient->isConnected()) {
		theTrayIcon->setPixmap(*On);
	} else {
		theTrayIcon->setPixmap(*Off);
	}
}

void KAnyRemote::resetCfgFromGui()
{
	theLog->debug("KAnyRemote::resetCfgFromGui()");
	cfgFileName = cfgWin->fileName->url().ascii();
	
	saveToConfig();	

	resetCfg(true);
	
	theClient->setStyle(conf->getDevice());
	commander->setStyle(theClient->isServer());
	
	if(theClient->isServer()) {
	        theLog->debug("KAnyRemote::resetCfgFromGui() theClient->isServer()");
		if (!theClient->isConnected()) {
			theLog->debug("KAnyRemote::resetCfgFromGui() STATE_DEFSTYLE");
			state = STATE_DEFSTYLE;
		} 
	}
        
	theLog->debug(QString("KAnyRemote::resetCfgFromGui() ....contextMenu()->setItemVisible() ")+(!theClient->isServer()?"TRUE":"FALSE"));
	theTrayIcon->contextMenu()->setItemVisible (menuItemConnectID, (!theClient->isServer()));
}

// cfgFileName must be set before
void KAnyRemote::resetCfg(bool printConf)
{
	//theLog->debug(QString("KAnyRemote::resetCfg() ") + cfgFileName);

	QString oldDevice = conf->getDevice();

	if (conf) {
		commander->freeTimers();
		conf->freeCfg();
		
		conf->init_cfg(cfgFileName);

		if(!conf->getLog() && theLog) {
			theLog->block(true);
		}
		theLog->setFile(conf->getLog(), conf->getDebug());
	} else {
		theLog->error("Configuration object is absent ?!");
		return;
	}

	if (theResetCount >= 1) {
		KPassivePopup::message("KAnyRemote", i18n("Resetting configuration settings."),
					*On, 
					theTrayIcon);

		// Unconnect if device in new cfg file differs from old
		if (oldDevice != conf->getDevice()) {	// change it despite connected or not
			//theLog->debug("KAnyRemote::resetCfg() new device is "+conf->getDevice());
			
			theClient->Disconnect();
			
			if (conf->getAutoConnect() || theClient->isServer()) {
				QTimer::singleShot(1000, this, SLOT(checkBluetooth()));
			}
			setPixmapsAndTooltips(0);
		}
	}

	if (theResetCount == 0) {
		if(conf->getLog() && theLog) { 
			theLog->truncate();
		}
	}
	
	if(printConf) {
	    conf->printConf();
	}
	theResetCount++;
}

void KAnyRemote::doQuit()
{
	theLog->debug("KAnyRemote::doQuit()");
	//KSimpleConfig theConfig("kanyremoterc");
	//theConfig.setGroup("General");
	QString cmd(getenv("HOME"));
	
	switch(KMessageBox::questionYesNoCancel(0, i18n("Should the KAnyRemote start automatically when you begin KDE?"), 
					i18n("Automatically Start?"), i18n("Start Automatically"), i18n("Do Not Start")))
	{	
		//case KMessageBox::No: theConfig.writeEntry("AutoStart", false); break;
		//case KMessageBox::Yes: theConfig.writeEntry("AutoStart", true); break;
		case KMessageBox::No:
			cmd =  "rm -f " + cmd + "/.kde/Autostart/kanyremote";
			theLog->debug("Autostart "+cmd);
			commander->execInFork(cmd.ascii());
			break;
		case KMessageBox::Yes: 
			cmd =  "ln -s `which kanyremote` " + cmd + "/.kde/Autostart/kanyremote";
			theLog->debug("Autostart "+cmd);
			commander->execInFork(cmd.ascii());
			break;
		case KMessageBox::Cancel: return;
	}
	
	//theConfig.writeEntry("AutoStart", false);
        
	commander->freeTimers();
	conf->freeCfg();
	
	KApplication::kApplication()->quit();
}

// If daemon - try to connect by timeout
void KAnyRemote::checkBluetooth()
{
	theLog->debug("KAnyRemote::checkBluetooth()");
	if (theClient->isConnected()) {
		theLog->debug("KAnyRemote::checkBluetooth() already connected !");
	} else {
		if (theClient->isServer()) {	    // Server-style
			if (state ==  STATE_DEFSTYLE) {
				if(theClient->initServerPort()) {
					theLog->debug("KAnyRemote::checkBluetooth() bind successful");
					state = STATE_SR_BIND;
					theClient->setInitialized();
				} else {
					theLog->error("KAnyRemote::checkBluetooth() can not bind !");
				}
			} else if (state == STATE_SR_BIND) {
				if (theClient->listenForConnection() == 1) {
					theLog->debug("KAnyRemote::checkBluetooth() connection established");
					state = STATE_SR_READ;
		
					setPixmapsAndTooltips(1, "Connected to phone");
		
					checkAlarms();
		
					QTimer::singleShot(200, this, SLOT(checkData()));
					return;
				// this is normal, since we did not establish connection yet
				//} else {
				//	theLog->error("KAnyRemote::checkBluetooth() fails in listen() or in accept()");
				}
			} else {
				theLog->debug("KAnyRemote::checkBluetooth() status inappropriate");
				return;
			}
			QTimer::singleShot(2000, this, SLOT(checkBluetooth()));

		} else {			    // AT-style
			theLog->debug("not connected yet ...");
			slotConnect();

			if ((!theClient->isConnected()) && conf->getAutoConnect()) {
				QTimer::singleShot(30000, this, SLOT(checkBluetooth()));
			}
		}
	}
}

// Manual (without timers) connect/disconnect + GUI stuff. Used only in client style
void KAnyRemote::Connect()
{
	theLog->debug("KAnyRemote::Connect()");
    	if(theClient->isServer()) {
	    theLog->debug("KAnyRemote::Connect() can not be used in server style");
	    return;
	}
	
	// Flash icon
	gotMessage("");

	if(theClient->isConnected()) {
		// Then disconnect
		theClient->Disconnect();
                theTrayIcon->contextMenu()->changeItem(menuItemConnectID, SmallIcon( "connect_creating" ), i18n("&Connect..."));
		setPixmapsAndTooltips(0);
	} else {
		// Then connect
		if(theClient->connectToBT(conf->getDevice())) {
			theTrayIcon->contextMenu()->changeItem (menuItemConnectID, SmallIcon( "connect_no" ), i18n("&Disconnect..."));
			setPixmapsAndTooltips(1, "Connected to phone");
			
			checkAlarms();

			theLog->debug("KAnyRemote::Connect() start checkData()");
			QTimer::singleShot(200, this, SLOT(checkData()));
		} else {
                	theTrayIcon->contextMenu()->changeItem(menuItemConnectID, SmallIcon( "connect_creating" ), i18n("&Connect..."));
                        setPixmapsAndTooltips(0);
		}
	} 
}

void KAnyRemote::slotConnect()
{
	Connect();
}

// Check alarm files and send message to the BT device if any
void KAnyRemote::checkAlarms()
{
	//theLog->info("KAnyRemote::checkAlarms()");
	theClient->manageAlarms(ALARM_CHECK);
	
	// Set it again only if connected
	if(theClient->isConnected()) {
		QTimer::singleShot(15000, this, SLOT(checkAlarms()));
	}
}

void KAnyRemote::handleExitAndDisc(int ret)
{
	theLog->debug("KAnyRemote::handleExitAndDisc()");

   	if (ret == EXIT_ABORT) {
    		theLog->info("User asks to exit");
		doQuit();
    	} else if (ret == EXIT_DISCON) {
		theLog->info("User asks to temporary disconnect");

    		// Already disconnected
		setPixmapsAndTooltips(0, i18n("User asks for temporary disconnect..."));

		// Connect after 1 minute
		QTimer::singleShot(60000, this, SLOT(checkBluetooth()));
	}
}

// AutoRepeat of pressed keys
void KAnyRemote::autoRepeat()
{
	theLog->info("KAnyRemote::autoRepeat()");
	
	// Set it again only if connected
	if(theClient->isConnected() && theClient->stillPressed() && theRepeatCounter < 100) {
		
		handleExitAndDisc(commander->handleCmdByKey(theClient->getAutoRepeatCmdIndex(),NULL));
		theRepeatCounter++;
		
		// 10 times per second 
		QTimer::singleShot(100, this, SLOT(autoRepeat()));
	} 
}

// Do we need to execute command by timer and send results to client ?
void KAnyRemote::handleTimerCmd()
{
	//theLog->debug("KAnyRemote::handleTimerCmd()");

	// Set it again only if connected
	if(theClient->isConnected()) {
		if (commander->verifyTimers(1) > 0) {  // There are exists active timers
			// 2 times per second 
			QTimer::singleShot(500, this, SLOT(handleTimerCmd()));
		}
	//} else {
	//  theLog->debug("KAnyRemote::handleTimerCmd() disconnected, do not reset timer");
	}
}

void KAnyRemote::getCallerID()
{
	theLog->debug("KAnyRemote::getCallerID()");

	// Set it again only if connected
	if(theClient->checkCallActive() == EXIT_OK && theClient->isConnected()) {
		// 1 time per second 
		QTimer::singleShot(1000, this, SLOT(getCallerID()));
	}
}

// Check presence of input data from BT device + handling
void KAnyRemote::checkData()
{
	QString answer;
	int ret;
	//theLog->debug("KAnyRemote::checkData()");

	if (theClient->isConnected()) {
		ret = theClient->readCommand(answer, 1024);
		if (ret == EOF) {
			theLog->info("Got disconnected");
		
			bool needClose = true;
			if (theClient->isServer())  {	// In server style we will reuse already opened socket
				needClose = false;
			}
			theClient->handleDisconnect(needClose);
			setPixmapsAndTooltips(0);

			if (conf->getAutoConnect() || theClient->isServer()) {
				theLog->debug("Set timer to auto connect");
				state = STATE_SR_BIND;
				QTimer::singleShot(5000, this, SLOT(checkBluetooth()));
		
				return;
			}
		} else if (ret) {
			//theLog->debug(QString("Command read >") + answer + "<");
		
			if (!(answer.isNull() || answer.isEmpty())) {
				gotMessage(answer);
		
				ret = theClient->handleCommand((char *)answer.ascii());
				if (ret == EXIT_OK) {
					theLog->debug("Processed... OK");
				} else {
					handleExitAndDisc(ret);
				}
			//} else {
			//  theLog->debug("Empty data was read");
			}
		}
		// In fact it could became disconnected in theClient->handle_command()
		if (theClient->isConnected()) {
			// 20 times per second
			QTimer::singleShot(50, this, SLOT(checkData()));
		}
	} else {
		theLog->debug("BT disconnected. Skip ...");
	}
}

// GUI way to set cfg file name
void KAnyRemote::slotConfigure()
{
	theLog->debug(QString("KAnyRemote::slotConfigure() ") + cfgFileName);

	cfgWin->fileName->clear();
	if (!(cfgFileName.isNull() || cfgFileName.isEmpty())) {
		cfgWin->fileName->setURL(cfgFileName);
	} 

	connect((const QObject *) cfgWin->ok, (const char *) SIGNAL( clicked() ), 
                 (const QObject *) this, (const char *) SLOT(resetCfgFromGui()));

	show();
}

// DCOP way to set cfg file name
void KAnyRemote::SetCfgFile(QString &theFile)
{
	theLog->debug("KAnyRemote::SetCfgFile()");
	cfgWin->fileName->setURL(theFile);
	resetCfgFromGui();
}

void KAnyRemote::loadFromConfig(void)
{
	//theLog->debug("KAnyRemote::loadFromConfig()");

	// load configuration from config file
	KSimpleConfig theConfig("kanyremoterc");
	theConfig.setGroup("General");
	cfgFileName = theConfig.readEntry("CfgFile");
	
	//theLog->cfg(QString("CfgFile=") + cfgFileName);
}

void KAnyRemote::saveToConfig(void)
{
	//theLog->debug("KAnyRemote::saveToConfig()");

	// load configuration from config file
	KSimpleConfig theConfig("kanyremoterc");

	theConfig.setGroup("General");
	theConfig.writeEntry("CfgFile", cfgFileName);
}

void KAnyRemote::showAbout()
{
	About->aboutApplication();
}

void KAnyRemote::resetRepeatCounter()
{
	theRepeatCounter = 0;
}
#include "kanyremote.moc"


